package page_vote

import (
	"code.electrolab.fr/it/vote.electrolab.fr/application"
	"code.electrolab.fr/it/vote.electrolab.fr/pages/common"
	"context"
	"database/sql"
	"fmt"
	"github.com/lib/pq"
	"html/template"
	"log"
	"math/rand"
	"net/http"
	"strings"
	"time"
)

const (
	FORM_VOTER_PRIVATE_TOKEN = "private_token"
)

var (
	templateChoice    = template.Must(template.New("index").Parse(TEMPLATE_CHOICE_STR))
	templateScore     = template.Must(template.New("index").Parse(TEMPLATE_SCORE_STR))
	templateForbidden = template.Must(template.New("forbidden").Parse(TEMPLATE_FORBIDDEN_STR))
)

func RegisterRoutes(mux *http.ServeMux, app *application.App) {
	mux.Handle("/votes/", application.NewAppHandler(app, handleGetVote))
}

func handleGetVote(app *application.App, ctx context.Context, w http.ResponseWriter, r *http.Request) {
	var (
		template     *template.Template
		templateVars = new(TemplateVars)
	)

	pathParts := strings.Split(r.URL.Path, "/")
	if len(pathParts) != 3 {
		w.WriteHeader(http.StatusNotFound)
		return
	}

	if userUuid, voterFullName, err := common.GetVoterInfo(app, ctx, r); err != nil {
		log.Println(err.Error())
		// Token is not valid
		w.WriteHeader(http.StatusForbidden)
		templateForbidden.Execute(w, nil)
		return
	} else {
		templateVars.VoterPrivateToken = r.FormValue("private_token")
		templateVars.UserUuid = userUuid
		templateVars.VoterFullname = voterFullName
	}

	questionName := pathParts[2]
	err := app.DB.QueryRowContext(ctx, `SELECT name, title, position, sentence, min_choices, COALESCE(max_choices, -1), vote_count, vote_open FROM questions WHERE name = $1 ORDER BY position`, questionName).Scan(&templateVars.Name, &templateVars.Title, &templateVars.Position, &templateVars.Sentence, &templateVars.MinChoices, &templateVars.MaxChoices, &templateVars.VoteCount, &templateVars.VoteOpen)
	if err != nil {
		if err == sql.ErrNoRows {
			// Question is not found
			// w.Header().Set("Location", "/votes")
			w.WriteHeader(http.StatusNotFound)
			return
		} else {
			// Query has an error
			log.Printf("Query error 1: %s", err.Error())
			w.WriteHeader(http.StatusInternalServerError)
			return
		}
	}

	if r.Method == http.MethodPost {
		err := doVote(app, ctx, r, templateVars)
		if err != nil {
			templateVars.VoteError = err.Error()
			fmt.Println("vote error", templateVars.VoteError)
		} else {
			// Vote is a success, redirect to the page as GET
			w.Header().Set("Location", r.URL.String())
			w.WriteHeader(http.StatusMovedPermanently)
			return
		}
	}

	rows, err := app.DB.QueryContext(ctx, `SELECT name, checked, sentence, vote_count FROM answers WHERE question_name = $1 ORDER BY position`, questionName)
	if err != nil {
		// Query has an error
		log.Printf("Query error 2: %s", err.Error())
		w.WriteHeader(http.StatusInternalServerError)
		return
	}
	defer rows.Close()

	for rows.Next() {
		var answer TemplateAnswer

		err = rows.Scan(&answer.Name, &answer.Checked, &answer.Sentence, &answer.VoteCount)
		if err != nil {
			log.Printf("Query error 3: %s", err.Error())
			w.WriteHeader(http.StatusInternalServerError)
			return
		}

		templateVars.Answers = append(templateVars.Answers, answer)
	}

	err = app.DB.QueryRowContext(ctx, `SELECT COALESCE((SELECT TRUE FROM votes WHERE question_name = $1 AND user_uuid = $2), FALSE)`, templateVars.Name, templateVars.UserUuid).Scan(&templateVars.VotedAlready)
	if err != nil {
		// Query has an error
		log.Printf("Query error 7: %s", err.Error())
		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	if templateVars.VoteOpen || templateVars.VoteCount == 0 {
		template = templateChoice
		shuffleAnswers(templateVars.Answers)
	} else {
		// If the vote is closed, and there is at least one recorded vote, then consider the vote done,
		// and switch to the 'score' template to show the results
		template = templateScore
	}

	err = template.Execute(w, templateVars)
	if err != nil {
		// Template has an error
		log.Printf("Template execution error: %s", err.Error())
		w.WriteHeader(http.StatusInternalServerError)
		return
	}
	return
}

// Pseudorandom shuffle to prevent order bias in answers
func shuffleAnswers(answers []TemplateAnswer) {
	rnd := rand.New(rand.NewSource(time.Now().UnixNano()))
	for i := len(answers) - 1; i > 0; i-- {
		j := rnd.Intn(i + 1)
		answers[i], answers[j] = answers[j], answers[i]
	}
}

func doVote(app *application.App, ctx context.Context, r *http.Request, templateVars *TemplateVars) error {
	var voteErrorMsg string

	err := app.DB.QueryRowContext(ctx,
		"SELECT COALESCE(vote($1, $2, $3), '')",
		templateVars.Name,
		templateVars.VoterPrivateToken,
		pq.StringArray(r.Form["answer"]),
	).Scan(&voteErrorMsg)

	if err != nil {
		return fmt.Errorf("SQL query error: %s", err.Error())
	}

	if len(voteErrorMsg) > 0 {
		return fmt.Errorf("%s", voteErrorMsg)
	}

	return nil
}
